@@ -8,17 +8,20 @@ use Solaris::Disk::Mnttab;
 use Term::ANSIColor;
 
 my $VERSION;
-$VERSION = 0.01;
+$VERSION = 0.02;
 
 =head1 NAME
 
-Salaris::Disk::SVM
+Solaris::Disk::SVM - Read, explore and manipulate SVM disk configurations
 
 =head1 SYNOPSIS
 
   my $svm = Solaris::Disk::SVM->new( init => 1 );
   $svm->dumpconfig();
 
+  my $svm = Solaris::Disk::SVM->new( init => 1, sourcedir => 't/data' );
+  $svm->dumpconfig();
+
   my $svm = Solaris::Disk::SVM->new( init => 0 );
 
   $svm->{mnttab}->readmtab( mnttab  => 't/data/mnttab.txt' );
@@ -74,13 +77,13 @@ Initialise and read tables, from optional sources:
 
 If running on different OS than Solaris, or on a different host than the target
 one, you may need to specify other data sources such as the one for
-C<Solaris::Disk::Mnttab> and C<Solaris::Disk::Partitions>.
+C<Solaris::Disk::Mnttab> and C<Solaris::Disk::VTOC>.
 
 =cut
 
 sub new {
     my ( $class, @args ) = @_;
-    $class = ref($class) || $class;
+#    $class = ref($class) || $class;
 
     my $self = {};
 
@@ -99,7 +102,7 @@ sub new {
     }
 
     # shouldn't be anything left in # @args
-    croak "Unknown parameter(s): @args" if @args;
+    warn "Unknown parameter(s): @args" if @args;
 
     # Create the vtoc object, but do initialize it.
     # In case of file based source, it would be initialized explicitely, and
@@ -110,14 +113,21 @@ sub new {
     $self->{mnttab} = Solaris::Disk::Mnttab->new( init => 0 );
     bless $self, $class;
 
+    if (defined $parms{sourcedir}) {
+        $parms{metastatp} = $parms{sourcedir}. "/metastat-p.txt";
+        $self->{mnttab}->readmtab( mnttab => $parms{sourcedir}."/mnttab.txt" );
+        $self->{mnttab}->readstab( swaptab => $parms{sourcedir}. "/swaptab.txt" );
+        $self->{vtoc}->readvtocdir( $parms{sourcedir} );
+    }
+
     if ( defined( $parms{init} ) && $parms{init} == 1 ) {
         $self->readconfig(
             defined( $parms{metastatp} )
             ? ( metastatp => $parms{metastatp} )
-            : () );
+            : () ) ;
     }
 
-    return $self;
+    $self;
 }
 
 =head2 C<sdscreateobject>
@@ -136,6 +146,7 @@ sub sdscreateobject($$) {
     my @desc;
 
     if ( $dev =~ m/^d\d+$/ ) {
+#        print STDERR "Reading $dev\n";
         for ($desc) {
 
             # A soft partition
@@ -144,38 +155,51 @@ sub sdscreateobject($$) {
                 my %spdesc;
                 while ( my $opt = shift @desc ) {
                     if ( $opt eq '-p' ) {
-                        $spdesc{slice} = shift @desc;
+                        my $spd = shift @desc;
+                        $spdesc{device} = $spd
+                          if defined $spd;
                     }
-                    elsif ( $opt eq '-o' ) {
+                    if ( $opt eq '-o' ) {
                         my $offset = shift @desc;
-                        push @{ $spdesc{offsets} }, $offset;
-                        $opt = shift @desc;
-                        if ( $opt eq '-b' ) {
-                            push @{ $spdesc{sizes} }, shift @desc;
-                        }
-                        else {
-                            croak
-"Soft-partition $dev has no size at offset $offset";
-                        }
+                        push @{ $spdesc{offsets} }, $offset
+                          if defined $offset;
                     }
-                    else {
-                        croak "Soft-partition $dev has no offset";
+                    if ( $opt eq '-b' ) {
+                        my $blocks = shift @desc;
+                        push @{ $spdesc{sizes} }, $blocks
+                          if defined $blocks;
                     }
                 }
+                if (
+                    !(
+                           defined( $spdesc{sizes} )
+                        && defined( $spdesc{offsets} )
+#                        && defined( $spdesc{device} )
+                    )
+                  )
+                {
+                    warn "Incomplete soft partition definition for $dev";
+                    delete $self->{Metastat}{$dev};
+                    return;
+                }
+                
                 $self->{devices}{$dev}{type} = "softpart";
 
-            #            $self->{objtype}{$self->{devices}{$dev}{type}}{$dev}++;
                 $self->{objtype}{ $self->{devices}{$dev}{type} }{$dev}{device} =
-                  $spdesc{slice};
+                  $spdesc{device};
 
-                push @{ $self->{LeafPhysDevices}{$dev} }, $spdesc{slice};
-
-                #      push @{$SubElements{$dev}}, $spdesc{slice};
+                if ( $spdesc{device} =~ /c\d+t\d+d\d+s\d+/ ) {
+                    $self->{LeafPhysDevices}{$dev}{$spdesc{device}}++;
+                    $self->{Partitions}{ $spdesc{device} }{use}++;
+                    $self->{PhysDevices4Dev}{$dev}{$spdesc{device}}++;
+                } else {
+                    push @{ $self->{SubElements}{$dev} }, $spdesc{device};
+                }
 
                 $self->{SPSize}{$dev} = 0;
-                foreach my $i ( 0 .. $#{ @{ $spdesc{offsets} } } ) {
+                foreach my $i ( 0 .. scalar(@{$spdesc{offsets}})-1 ) {
                     my $size = @{ $spdesc{sizes} }[$i];
-                    push @{ $self->{SPContains}{ $spdesc{slice} } },
+                    push @{ $self->{SPContains}{ $spdesc{device} } },
                       [
                         @{ $spdesc{offsets} }[$i],
                         @{ $spdesc{sizes} }[$i],
@@ -187,17 +211,17 @@ sub sdscreateobject($$) {
 
                 $self->{devices}{$dev}{explanation} =
                     "$dev is a soft partition (dev="
-                  . join( " ", @{ $self->{LeafPhysDevices}{$dev} } )
+                  . join( " ", keys %{$self->{LeafPhysDevices}{$dev}} )
                   . ", size="
                   . $self->{SPSize}{$dev}
                   . ", offset={"
                   . join( ', ', @{ $self->{SPOffset}{$dev} } ) . "}) ";
-                if ( $spdesc{slice} =~ /c\d+t\d+d\d+s\d+/ ) {
-                    $self->{Partitions}{ $spdesc{slice} }{use}++;
-                }
+
+
                 last;
             };
 
+            # RAID 5
             /^\-r.*/ and do {
                 @desc = split /\s+/, $desc;
                 shift @desc;    # pass by the -r
@@ -214,20 +238,21 @@ sub sdscreateobject($$) {
                         };
                         /^c\d+t\d+d\d+s\d+/ and do {
                             $self->{Partitions}{$el}{'use'}++;
-                            push @{ $self->{LeafPhysDevices}{$dev} }, $el;
+                            $self->{LeafPhysDevices}{$dev}{$el}++;
+                            $self->{PhysDevices4Dev}{$dev}{$el}++;
                             push @elements, $el;
                             $columns++;
                             last;
                         };
-                        /-k$/ and do {
+                        /^\-k$/ and do {
                             $self->{devices}{$dev}{kflag}++;
                             last;
                         };
-                        /-i$/ and do {
+                        /^\-i$/ and do {
                             $self->{devices}{$dev}{iflag} = shift @desc;
                             last;
                         };
-                        /-o$/ and do {
+                        /^\-o$/ and do {
                             $self->{devices}{$dev}{oflag} = shift @desc;
                             last;
                         };
@@ -243,6 +268,7 @@ sub sdscreateobject($$) {
                 last;
             };
 
+            # Trans (journalised device)
             /^\-t.*/ and do {
                 @desc = split /\s+/, $desc;
                 shift;    # pass by the -t
@@ -262,6 +288,7 @@ sub sdscreateobject($$) {
                 last;
             };
 
+            # mirror
             /^\-m.*/ and do {
 
                 #  print "$dev is a mirror ($desc)\n";
@@ -279,13 +306,11 @@ sub sdscreateobject($$) {
 
                 $self->{objtype}{ $self->{devices}{$dev}{type} }{$dev}{sides} =
                   scalar @{ $self->{SubElements}{$dev} };
-                print STDERR "WARNING : "
-                  . $self->{devices}{$dev}{type}
-                  . " $dev has only one side\n"
-                  if ( @{ $self->{SubElements}{$dev} } == 1 && $self->{debug} );
+                  # print STDERR "WARNING : " . $self->{devices}{$dev}{type} . " $dev has only one side\n" if ( @{ $self->{SubElements}{$dev} } == 1 );
                 last;
             };
 
+            # device / concat / stripe
             /^[0-9]+.*/ and do {
 
                 #  print "$dev is a concat stripe ($desc)\n";
@@ -300,17 +325,20 @@ sub sdscreateobject($$) {
                       if ( $maxconcat < $cstripe );    # max allg
                     foreach my $j ( 1 .. $cstripe ) {
                         my $el = shift @desc;
-                        if ( $el eq '-i' ) {
-                            shift @desc;
-                            $el = shift @desc;
-                        }
+                        
                         for ($el) {
-                            /^d\d+/ and do {
+                            /^d\d+$/ and do {
+                                # TODO: check if $el is a softpart
+                                #if ( defined($self->{devices}{$el}{type}) &&
+                                #    $self->{devices}{$el}{type} ne 'softpart') {
+                                #    warn "$el is already a device, not a soft-partition"
+                                #}
                                 push @{ $self->{SubElements}{$dev} }, $el;
                                 last;
                             };
-                            /^c\d+t\d+d\d+s\d+/ and do {
-                                push @{ $self->{LeafPhysDevices}{$dev} }, $el;
+                            /^c\d+t\d+d\d+s\d+$/ and do {
+                                $self->{LeafPhysDevices}{$dev}{$el}++;
+                                $self->{PhysDevices4Dev}{$dev}{$el}++;
 
                                 #   push @{ $self->{SubElements}{$dev} }, $el;
                                 $self->{Partitions}{$el}{'use'}++;
@@ -322,14 +350,21 @@ sub sdscreateobject($$) {
                     }
                 }
 
+                my $el = shift @desc;
+                if (defined $el and $el eq '-i' ) {
+                    my $fl = shift @desc;
+                    $self->{devices}{$dev}{iflag} = $fl
+                      if defined $fl;
+                }
+
                 if ( $nstripe == 1 && $maxconcat == 1 ) {
                     $self->{devices}{$dev}{type} = "device";
                 }
                 if ( $nstripe > 1 && $maxconcat == 1 ) {
-                    $self->{devices}{$dev}{type} = "concat";
+                    $self->{devices}{$dev}{type} = "stripe";
                 }
                 if ( $nstripe == 1 && $maxconcat > 1 ) {
-                    $self->{devices}{$dev}{type} = "stripe";
+                    $self->{devices}{$dev}{type} = "concat";
                 }
                 if ( $nstripe > 1 && $maxconcat > 1 ) {
                     $self->{devices}{$dev}{type} = "concat+stripe";
@@ -344,24 +379,27 @@ sub sdscreateobject($$) {
             };
         }
     }
-    if ( $dev =~ m/^hsp\d+/ ) {
 
-        # FIXME : need to verify the hsp format
+    # hot-spare
+    if ( $dev =~ m/^hsp\d+/ ) {
         my @desc = split /\s+/, $desc;
 
-        $self->{SubElements}{$dev} = @desc ? @desc : ();
         $self->{devices}{$dev}{type} = 'hotspare';
         if (@desc) {
+            print "$dev\n";
+            $self->{SubElements}{$dev} = @desc;
             $self->{devices}{$dev}{explanation} =
               "$dev is a hotspare ("
-              . join( " ", @{ $self->{SubElements}{$dev} } ) . ")";
+              . join( " ", @desc) . ")";
+            $self->{objtype}{hotspare}{$dev}{disks} = scalar @desc;
         }
         else {
+            $self->{SubElements}{$dev} = ();
+            $self->{objtype}{hotspare}{$dev}{disks} = 0;
             $self->{devices}{$dev}{explanation} =
               "$dev is a hotspare without any disk";
 
         }
-        $self->{objtype}{hotspare}{$dev}{disks} = scalar @desc;
     }
 }
 
@@ -376,6 +414,8 @@ dependencies.
 The C<metastatp> named argument can provide a file source to use instead of
 'F<metastat -p |>'.
 
+Returns non 0 on error.
+
 =cut
 
 sub readconfig(@) {
@@ -400,16 +440,20 @@ sub readconfig(@) {
     }
 
     # shouldn't be anything left in @args
-    croak "Unknown parameter(s): @args"
+    warn "Unknown parameter(s): @args"
       if @args;
 
     $self->{metastatSource} = "metastat -p |";  # the default source... maybe
                                                 # should we specify full path...
 
-    $self->{metastatSource} = $parms{metastatp} if defined $parms{metastatp};
+    $self->{metastatSource} = $parms{metastatp}
+      if defined $parms{metastatp};
+
+    if (! open MS, $self->{metastatSource}) {
+        warn "Cannot open metastat source '$self->{metastatSource}'";
+        return -1;
+    }
 
-    open MS, $self->{metastatSource}
-      or croak "Cannot open metastat source '$self->{metastatSource}'";
     while (<MS>) {
         next if m/^#/;
         while ( $_ =~ /.*\\\n/ ) {
@@ -418,44 +462,52 @@ sub readconfig(@) {
             $_ .= readline(*MS);    # concat next line from MS
         }
         ( $dev, @desc ) = split;
-        $self->{Metastat}{$dev} = join " ", @desc;
+
+        next if not defined($dev);  # empty line
+
+        if ( defined $self->{Metastat}{$dev} ) {
+            warn "$dev is already defined, and one more definition is given";
+        }
+        else {
+            $self->{Metastat}{$dev} = join " ", @desc;
+            $self->sdscreateobject( $dev, $self->{Metastat}{$dev} );
+        }
+    }
+    if (! close MS) {
+        warn "lost metastat source '$self->{metastatSource}'";
+        return -1;
     }
-    close MS;
 
     # On a maintenant un hash %Metastat qui contient la config brute
-    while ( ( $dev, $desc ) = each %{ $self->{Metastat} } ) {
-        $self->sdscreateobject( $dev, $desc );
-    }
+#    while ( ( $dev, $desc ) = each %{ $self->{Metastat} } ) { $self->sdscreateobject( $dev, $desc ); }
 
     #
     # Passe pour assigner les physical  tt le monde
     #
     foreach $dev ( keys %{ $self->{Metastat} } ) {
         foreach my $sdev ( $self->getsubdevs($dev) ) {
-            if ( defined( @{ $self->{LeafPhysDevices}{$sdev} } ) ) {
-
-#                push @{ $self->{PhysDevices4Dev}{$dev} }, @{ $self->{LeafPhysDevices}{$sdev} };
-                foreach my $i ( @{ $self->{LeafPhysDevices}{$sdev} } ) {
-
-                 #                    print STDERR "$sdev is subdev for $dev\n";
+            if ( defined( %{ $self->{LeafPhysDevices}{$sdev} } ) ) {
+                foreach my $i ( keys %{ $self->{LeafPhysDevices}{$sdev} } ) {
                     $self->{PhysDevices4Dev}{$dev}{$i}++;
                 }
             }
         }
-    }
+   }
 
     # FIXME : test avant de lancer le readvtoc
     #
     # Recuperation des informations des partitions
     #
     foreach $dev ( keys %{ $self->{Partitions} } ) {
-        if ( $dev =~ m/c\d+t\d+d\d+/ ) {
+#        if ( $dev =~ m/c\d+t\d+d\d+/ ) {
             $self->{vtoc}->readvtoc( device => $dev );
             $self->{Partitions}{$dev}{size} = $self->{vtoc}->size($dev);
-        }
+#        }
     }
     foreach $dev ( keys %{ $self->{LeafPhysDevices} } ) {
-        $self->{vtoc}->readvtoc( device => $dev ) if $dev =~ m/c\d+t\d+d\d+/;
+        foreach my $part (keys %{$self->{LeafPhysDevices}{$dev}}) {
+            $self->{vtoc}->readvtoc( device => $part );
+        }
     }
 
     # A priori, les informations sur les partitions sont maintenant compltes,
@@ -467,15 +519,13 @@ sub readconfig(@) {
         next if $dev !~ m/(?:d\d+|c\d+t\d+d\d+(?:s\d+)*)/;
         my $mnt = $self->{mnttab}->{dev2mp}{$dev};
 
-        #print STDERR "mnt = $mnt\n";
         # Copie dans l'objet svm
-        $self->{dev2mp}{$dev} = $mnt;
+        $self->{dev2mp}{$dev}{$mnt}++;
         $self->{mp2dev}{$mnt} = $dev;
-        foreach my $sdev ( $self->getsubdevs($dev) ) {
-            $self->{dev2mp}{$sdev}{$mnt}++;
+        foreach my $subdev ( $self->getsubdevs($dev) ) {
+            $self->{dev2mp}{$subdev}{$mnt}++;
         }
         foreach my $pdev ( keys %{ $self->{PhysDevices4Dev}{$dev} } ) {
-
             #           print STDERR "phys dev $pdev\n";
             $self->{pdev2mp}{$pdev}{$mnt}++;
         }
@@ -500,13 +550,13 @@ sub size($$) {
 
     if ( $dev =~ m/(c\d+t\d+d\d+)s(\d+)/ ) {
         ( $disk, $pn ) = ( $1, 'slice' . $2 );
-        $size = $self->{vtoc}{$disk}{$pn}{'count'};
+        $size = $self->{vtoc}{$disk}{$pn}{count};
     }
     elsif ( $dev =~ m/d\d+/ ) {
         if ( defined( $self->{SPSize}{$dev} ) ) {
             $size = $self->{SPSize}{$dev};
         }
-        elsif ( scalar @{ $self->{SubElements}{$dev} } ) {
+        elsif ( defined($self->{devices}{$dev}) && scalar @{ $self->{SubElements}{$dev} } ) {
             my $Msize = 0;
             my $msize = 1 << 31;    # some maximum value
             foreach my $se ( @{ $self->{SubElements}{$dev} } ) {
@@ -518,25 +568,23 @@ sub size($$) {
             if ( $self->{devices}{$dev}{type} eq 'trans' ) {
                 $size = $Msize;    # le maxima
             }
-            elsif ( $self->{devices}{$dev}{type} eq 'mirror' ) {
+            if ( $self->{devices}{$dev}{type} eq 'mirror' ) {
                 $size = $msize;    # le minima
             }
-            elsif ( $self->{devices}{$dev}{type} eq 'raid5' ) {
+            if ( $self->{devices}{$dev}{type} eq 'raid5' ) {
                 $size = $msize * ( scalar(@{ $self->{SubElements}{$dev} }) - 1 );
             }
         }
-        elsif ( scalar @{ $self->{LeafPhysDevices}{$dev} } ) {
+        elsif ( scalar keys %{$self->{LeafPhysDevices}{$dev}} ) {
             my $msize = 1 << 31;    # some maximum value
-            foreach my $se ( @{ $self->{LeafPhysDevices}{$dev} } ) {
+            foreach my $se ( keys %{ $self->{LeafPhysDevices}{$dev} } ) {
                 my $s = $self->{vtoc}->size($se);
                 $size += $s;
                 $msize = $s if $s < $msize;
             }
-            if ( $self->{devices}{$dev}{type} eq 'mirror' ) {
-                $size = $msize;    # le minima
-            }
-            elsif ( $self->{devices}{$dev}{type} eq 'raid5' ) {
-                $size = $msize * ( scalar(@{ $self->{LeafPhysDevices}{$dev} }) - 1 );
+#            if ( $self->{devices}{$dev}{type} eq 'mirror' ) { $size = $msize;    # le minima }
+            if ( $self->{devices}{$dev}{type} eq 'raid5' ) {
+                $size = $msize * ( scalar(keys %{ $self->{LeafPhysDevices}{$dev} }) - 1 );
             }
         }
 
@@ -555,32 +603,42 @@ C<showconfig> dumps in a (almost) human readable manner the configuration.
 
 sub showconfig($) {
     my ($self) = @_;
-    print "Apercu de la configuration :\n";
-    my @devs = keys %{ $self->{devices} };
 
-    foreach (@devs) {
-        s/^d//;
+    my (@devs, @hsps);
+
+    foreach (keys %{ $self->{devices} } ) {
+            push( @devs, $1) if m/^d(\d+)/;
+            push( @hsps, $1) if m/^hsp(\d+)/;
     }
 
+    print "SVM Configuration:\n";
     foreach ( sort { $a <=> $b } @devs ) {
-        my $devsize = $self->size( 'd' . $_ ) >> 11;
-        print "d$_ ("
-          . $self->{devices}{ 'd' . $_ }{type} . ") : "
-          . $self->{devices}{ 'd' . $_ }{explanation}
-          . "($devsize Mo)\n";
+        my $dev = "d$_";
+        my $devsize = $self->size($dev) >> 11;
+        print "$dev ("
+          . $self->{devices}{$dev}{type} . "): "
+          . $self->{devices}{$dev}{explanation}
+          . " [$devsize Mo]\n";
+    }
+
+    foreach ( sort { $a <=> $b } @hsps ) {
+        my $dev = "hsp$_";
+        print "$dev ("
+          . $self->{devices}{$dev}{type} . "): "
+          . $self->{devices}{$dev}{explanation}
+          . "\n";
     }
 }
 
 =head2 C<dumpconfig>
 
-C<dumpconfig> permettra de sortir la configuration au format metastat -p.
+C<dumpconfig> will, when implemented, dump the loaded configuration in the
+"metastat -p" format.
 
 =cut
 
-sub dumpconfig {
+sub dumpconfig { 1; }
 
-    # FIXME
-}
 
 =head2 C<showsp>
 
@@ -594,7 +652,17 @@ C<showsp> prints the list of the given soft-partition(s) container.
 
 sub showsp(@) {
     my ( $self, @devs ) = @_;
-    sub _sortbyoffset { $a->[0] <=> $b->[0] }
+#    sub _sortbyoffset { $a->[0] <=> $b->[0] }
+
+    my ( $reset, $red, $green );
+    if ( $self->{colour} ) {
+        $reset = color('reset');
+        $red   = color('red');
+        $green = color('green');
+    } else {
+        $reset = $red  = $green = "";
+    }
+    
 
     @devs = keys %{ $self->{SPContains} } if @devs == 0;
     foreach my $dev ( sort @devs ) {
@@ -608,14 +676,14 @@ sub showsp(@) {
 "----------+--------+------------+------------+-----------------------------------\n";
             my $tsize = 0;
             foreach
-              my $list ( sort _sortbyoffset @{ $self->{SPContains}{$dev} } )
+              my $list ( sort { $a->[0] <=> $b->[0] } @{ $self->{SPContains}{$dev} } )
             {
                 my $offset   = $list->[0];
                 my $size     = $list->[1];
                 my $endblock = $offset + $size - 1;
 
                 if ( ( $offset - $precendblock ) > 2 ) {
-                    print color('green') if $self->{colour};
+                    print $green;
                     printf(
                         "%-9s | *FREE* | %10d | %10d | %10d (%5d Mo) | free\n",
                         $dev,
@@ -624,7 +692,7 @@ sub showsp(@) {
                         $offset - $precendblock - 3,
                         ( $offset - $precendblock - 3 ) >> 11,
                     );
-                    print color('reset') if $self->{colour};
+                    print $reset;
                 }
 
                 printf(
@@ -649,37 +717,40 @@ sub showsp(@) {
                 if (
                     (
                         my $free =
-                        $self->{vtoc}->{$disk}{$pn}{'count'} -
+                        $self->{vtoc}->{$disk}{$pn}{count} -
                         ( $precendblock + 2 )
                     ) > 0
                   )
                 {
-                    print color('green') if $self->{colour};
+                    print $green;
                     printf(
                         "%-9s | *FREE* | %10d | %10d | %10d (%5d Mo) | free\n",
                         $dev,
                         $precendblock + 2,
-                        $self->{vtoc}->{$disk}{$pn}{'count'},
+                        $self->{vtoc}->{$disk}{$pn}{count},
                         $free, $free >> 11,
                     );
-                    print color('reset') if $self->{colour};
+                    print $reset;
                 }
                 else {
-                    print color('red') if $self->{colour};
+                    print $red;
                     printf( "%-9s | No more free space...\n", $dev );
-                    print color('reset') if $self->{colour};
+                    print $reset;
                 }
             }
             else {
                 my $devsize = $self->size($dev);
                 if ( ( my $free = $devsize - $precendblock - 1 ) > 0 ) {
-                    print color('green') if $self->{colour};
+                    print $green;
                     printf(
                         "%-9s | *FREE* | %10d | %10d | %10d (%5d Mo) | free\n",
                         $dev, $precendblock + 2,
                         $devsize, $free, $free >> 11,
                     );
-                    print color('reset') if $self->{colour};
+                    print $reset;
+                }
+                else {
+                    warn "More devices defined than free space available: should not happen";
                 }
             }
             print
@@ -704,7 +775,8 @@ sub explaindev(@) {
     my ( $self, @devs ) = @_;
 
     foreach my $dev (@devs) {
-        my $devsize = getsize($dev) >> 11;
+        next if not defined $self->{devices}{$dev};
+        my $devsize = $self->size($dev) >> 11;
         print $self->{devices}{$dev}{explanation}
           . ". It is $devsize MB large\n";
     }
@@ -724,22 +796,17 @@ Used to find what number to give to a new device.
 sub getnextdev($) {
     my ($self) = @_;
 
-    my @devs = keys %{ $self->{devices} };
+    # search among devices only
+    my @devs = sort { $a <=> $b } map { $1 if m/^d(\d+)/ } keys %{ $self->{devices} };
 
-    foreach (@devs) {
-        s/^d//;
-    }
-
-    my @sdev = sort { $a <=> $b } @devs;
-
-    $sdev[-1] + 1;
+    $devs[-1] + 1;
 }
 
 =head2 C<isdevfree>
 
   my $isfree{$dev} = $svm->isdevfree( $dev );
 
-C<isdevfree> take a device name (/d\d/) or a device number in argument
+C<isdevfree> take a device name (C</^d\d$/>) or a device number in argument
 and return 0 if the device is already defined, 1 if the device is not
 defined.
 
@@ -748,14 +815,9 @@ defined.
 sub isdevfree($$) {
     my ( $self, $dev ) = @_;
 
-    $dev = 'd' . $dev if ( $dev =~ /\d/ );
-    if ( exists $self->{Metastat}{$dev} ) {
-        explaindev($dev)             if $self->{verbose};
-        return 0;
-    }
-    else {
-        return 1;
-    }
+    $dev = "d$dev" if ( $dev =~ /^\d+$/ );
+#    if ( defined $self->{Metastat}{$dev} ) { return 0; } else { return 1; }
+    defined $self->{Metastat}{$dev} ? 0 : 1;
 }
 
 # $1 = dev_id
@@ -837,9 +899,8 @@ sub getsubdevs {
         if ( defined( @{ $self->{SubElements}{$dev} } ) ) {
             @subdevs = @{ $self->{SubElements}{$dev} };
             foreach my $cdev (@subdevs) {
-                if ( $cdev =~ /d\d+/ ) {
-                    push @subdevs, $self->getsubdevs($cdev);
-                }
+                push @subdevs, $self->getsubdevs($cdev);
+#                  if ($cdev =~ /d\d+/);
             }
         }
     }
@@ -859,13 +920,14 @@ passed as argument(s).
 
 sub getphysdevs(@) {
     my ( $self, @devs ) = @_;
-    my @pdevs;
-    foreach (@devs) {
-        if ( exists( $self->{Metastat}{$_} ) ) {
-            push @pdevs, keys %{ $self->{PhysDevices4Dev}{$_} };
+    my @pdevs = ();
+
+    foreach my $dev (@devs) {
+        if ( exists($self->{'Metastat'}{$dev} ) ) {
+            push @pdevs, keys %{ $self->{PhysDevices4Dev}{$dev} };
+#                if exists $self->{PhysDevices4Dev}{$dev};
         }
     }
-
     @pdevs;
 }
 
@@ -885,7 +947,6 @@ sub mponslice($$) {
     my ( $self, $slice ) = @_;
     my @mps;
 
-    $slice = $1 if $slice =~ m/(c\d+t\d+d\d+s\d+)/;
     if ( defined( $self->{pdev2mp}{$slice} ) ) {
         @mps = keys %{ $self->{pdev2mp}{$slice} };
     }
@@ -905,17 +966,17 @@ C<mpondisk> returns the list of filesystems present on a physical disk.
 
 sub mpondisk($$) {
     my ( $self, $disk ) = @_;
-    my @mps;
+    my %mps;
 
     $disk = $1 if $disk =~ m/(c\d+t\d+d\d+)/;
     foreach my $i ( 0 .. 7 ) {
         my $slice = "${disk}s$i";
         if ( defined( $self->{pdev2mp}{$slice} ) ) {
-            push @mps, keys %{ $self->{pdev2mp}{$slice} };
+            $mps{$_}++ foreach (keys %{ $self->{pdev2mp}{$slice} });
         }
     }
 
-    @mps;
+    keys %mps;
 }
 
 =head2 C<mpondev>
@@ -930,11 +991,11 @@ sub mpondev($$) {
     my ( $self, $dev ) = @_;
     my @mps;
 
-    $dev = $1 if $dev =~ m/(d\d+)/;
     if ( defined( $self->{dev2mp}{$dev} ) ) {
-        @mps = keys %{ $self->{dev2mp}{$dev} };
+        @mps = keys %{
+            $self->{dev2mp}{$dev}
+        };
     }
-
     @mps;
 }
 
@@ -952,7 +1013,8 @@ sub devs4mp($$) {
     my @devs;
 
     if ( defined( $self->{mp2dev}{$mp} ) ) {
-        @devs = $self->{mp2dev}{$mp}, $self->getsubdevs( $self->{mp2dev}{$mp} );
+        @devs =
+          ( $self->{mp2dev}{$mp}, $self->getsubdevs( $self->{mp2dev}{$mp} ) );
     }
 
     @devs;
@@ -970,12 +1032,11 @@ point via SVM.
 sub disks4mp($$) {
     my ( $self, $mp ) = @_;
     my @disks;
-
     if ( defined( $self->{mp2dev}{$mp} ) ) {
-        @disks = $self->getphysdevs( $self->{mp2pdev}{$mp} );
+        @disks = $self->getphysdevs( $self->{mp2dev}{$mp} );
     }
 
-    @disks;
+    return @disks;
 }
 
 =head2 C<version>
@@ -1001,34 +1062,46 @@ the lost in data structure programer syndrom(tm).
 RAID0 device size may not be accurate, particularly when the underlying
 devices are of different sizes. Your mileage may vary. This particularity needs
 testing and development in regard to the computing model (smallest device size *
-number of devices for RAID1.
+number of devices for RAID1).
 
-The test suite is quite limited as to possible SVM configuration. I really
-should take time to create weird configurations, with space loss as in concat
-stripes with different size components.
+The test suite has been augmented to cover all possible SVM configuration.
+However, these configurations may not be possible with SVM (I do not have access
+anymore to a Solaris+SVM machine), thus the module may accept unacceptable
+configuration schemes.
 
-Another good test to implement is a simple C<showconfig>, and see if there are
-no 'Use of uninitialized value' warnings. Seems I need to re-read Schwern's
-testing tutorial.
+I really should take time to create weird configurations, with space loss as in
+concat stripes with different size components, and users, such as you, will help
+all of us sending me some weird configurations as well. The more data we have to
+test against, the more accurate the module will be. If nearly all code pathes
+are actually tested, some are better than others. One example is the C<size>
+method, which results are B<not> tested against real world figures.
+
+Please report any other bugs or feature requests to
+C<bug-solaris-disk-svm@rt.cpan.org>, or through the web interface at
+L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=Solaris-Disk-SVM>.
+I will be notified, and then you'll automatically be notified of progress on
+your bug as I make changes.
 
 =head1 AUTHOR
 
 Jrme Fenal <jfenal@free.fr>
 
+You are welcome to send me sample configurations, bug reports or kudos.
+
 =head1 VERSION
 
-This is version 0.01 of the C<Solaris::Disk::SVM>.
+This is version 0.02 of the C<Solaris::Disk::SVM> module.
 
 =head1 COPYRIGHT
 
-Copyright (C) 2004 Jrme Fenal. All Rights Reserved
+Copyright (C) 2004, 2005 Jrme Fenal. All Rights Reserved
 
 This module is free software; you can redistribute it and/or modify it under the
 same terms as Perl itself.
 
-
 =head1 SEE ALSO
 
-See L<Solaris::Disk::Partitions> to access slice information.
+See L<Solaris::Disk::VTOC> to access disk slices information.
+
 See L<Solaris::Disk::Mnttab> to get current mounted devices.
 
